import re
import sys
from os import environ, getcwd, listdir, makedirs, rename, rmdir, unlink
from os.path import abspath, basename, dirname, exists, isdir, isfile, join, normpath, split, islink, lexists
from subprocess import check_output, CalledProcessError, STDOUT
import shutil

import jinja2
import yaml
from six import string_types
from concurrent.futures import Future, Executor
from threading import Lock

import future.builtins

from conda_verify.constants import MAGIC_HEADERS, DLL_TYPES

try:
    from functools import lru_cache
except ImportError:
    from backports.functools_lru_cache import lru_cache


@lru_cache(maxsize=32)
def yamlize(data):
    res = yaml.load(data)
    # ensure the result is a dict
    if res is None:
        res = {}
    return res


def ns_cfg(cfg):
    plat = "-".join((cfg["platform"], cfg["arch"]))
    py = int("".join(cfg["python"].split(".")))
    np = int("".join(cfg["numpy"].split(".")))

    return dict(
        nomkl=False,
        debug=False,
        linux=plat.startswith("linux-"),
        linux32=bool(plat == "linux-32"),
        linux64=bool(plat == "linux-64"),
        armv7l=bool(plat == "linux-armv7l"),
        arm=bool(plat == "linux-armv7l"),
        ppc64le=bool(plat == "linux-ppc64le"),
        osx=plat.startswith("osx-"),
        unix=plat.startswith(("linux-", "osx-")),
        win=plat.startswith("win-"),
        win32=bool(plat == "win-32"),
        win64=bool(plat == "win-64"),
        x86=plat.endswith(("-32", "-64")),
        x86_64=plat.endswith("-64"),
        py=py,
        py3k=bool(30 <= py < 40),
        py2k=bool(20 <= py < 30),
        py26=bool(py == 26),
        py27=bool(py == 27),
        py33=bool(py == 33),
        py34=bool(py == 34),
        py35=bool(py == 35),
        py36=bool(py == 36),
        np=np,
    )


sel_pat = re.compile(r"(.+?)\s*\[(.+)\]$")


def select_lines(data, namespace):
    lines = []
    for line in data.splitlines():
        line = line.rstrip()
        m = sel_pat.match(line)
        if m:
            cond = m.group(2)
            if eval(cond, namespace, {}):
                lines.append(m.group(1))
            continue
        lines.append(line)
    return "\n".join(lines) + "\n"


def parse(data, cfg):
    if cfg is not None:
        data = select_lines(data, ns_cfg(cfg))
    # ensure we create new object, because yamlize is memoized
    return dict(yamlize(data))


def render_jinja2(recipe_dir):
    loaders = [jinja2.FileSystemLoader(recipe_dir)]
    env = jinja2.Environment(loader=jinja2.ChoiceLoader(loaders))
    template = env.get_or_select_template("meta.yaml")
    return template.render(environment=env)


try:
    # circular dependency.  We only try this import so that conda-build is an optional
    #      conda-verify dep
    from conda_build import api

    def render_metadata(recipe_dir, cfg):
        m = api.render(
            recipe_dir, finalize=False, bypass_env_check=True, **(cfg if cfg else {})
        )[0][0]
        return m.get_rendered_recipe_text()


except ImportError:

    def render_metadata(recipe_dir, cfg):
        data = render_jinja2(recipe_dir)
        return parse(data, cfg)


def iter_cfgs():
    for py in "27", "34", "35":
        for plat in "linux-64", "linux-32", "osx-64", "win-32", "win-64":
            platform, arch = plat.split("-")
            yield dict(platform=platform, arch=arch, python=py, numpy="1.11")


def get_object_type(data):
    head = data[:4]
    if head not in MAGIC_HEADERS:
        return None
    lookup = MAGIC_HEADERS.get(head)
    if lookup == "DLL":
        pos = data.find(b"PE\0\0")
        if pos < 0:
            return "<no PE header found>"
        data = future.builtins.bytes(data)
        i = data[pos + 4] + 256 * data[pos + 5]
        return "DLL " + DLL_TYPES.get(i)
    elif lookup.startswith("MachO"):
        return lookup
    elif lookup == "ELF":
        return "ELF" + {"\x01": "32", "\x02": "64"}.get(data[4])


def get_bad_seq(s):
    for seq in ("--", "-.", "-_", ".-", "..", "._", "_-", "_."):  # but '__' is fine
        if seq in s:
            return seq
    return None


def all_ascii(data, allow_CR=False):
    newline = [10]  # LF
    if allow_CR:
        newline.append(13)  # CR
    for c in data:
        n = ord(c) if sys.version_info[0] == 2 else c
        if not (n in newline or 32 <= n < 127):
            return False
    return True


def ensure_list(argument):
    if isinstance(argument, list):
        return argument
    elif isinstance(argument, string_types):
        return argument.split(",")
    return [argument]


def fullmatch(regex, string, flags=0):
    """Emulate python-3.4 re.fullmatch().

    Credit: https://stackoverflow.com/questions/30212413/
    """
    return re.match("(?:" + regex + r")\Z", string, flags=flags)


# use this for debugging, because ProcessPoolExecutor isn't pdb/ipdb friendly
class DummyExecutor(Executor):
    def __init__(self):
        self._shutdown = False
        self._shutdownLock = Lock()

    def submit(self, fn, *args, **kwargs):
        with self._shutdownLock:
            if self._shutdown:
                raise RuntimeError("cannot schedule new futures after shutdown")

            f = Future()
            try:
                result = fn(*args, **kwargs)
            except BaseException as e:
                f.set_exception(e)
            else:
                f.set_result(result)

            return f

    def shutdown(self, wait=True):
        with self._shutdownLock:
            self._shutdown = True


#  Copied from conda because it is much faster and less deadlock-prone, but we don't want to include conda as a dep.

def which(executable):
    from distutils.spawn import find_executable
    return find_executable(executable)


def rmtree(path, *args, **kwargs):
    # subprocessing to delete large folders can be quite a bit faster
    path = normpath(path)
    if sys.platform == "win32":
        try:
            # the fastest way seems to be using DEL to recursively delete files
            # https://www.ghacks.net/2017/07/18/how-to-delete-large-folders-in-windows-super-fast/
            # However, this is not entirely safe, as it can end up following symlinks to folders
            # https://superuser.com/a/306618/184799
            # so, we stick with the slower, but hopefully safer way.  Maybe if we figured out how
            #    to scan for any possible symlinks, we could do the faster way.
            # out = check_output('DEL /F/Q/S *.* > NUL 2> NUL'.format(path), shell=True,
            #                    stderr=STDOUT, cwd=path)

            out = check_output('RD /S /Q "{}" > NUL 2> NUL'.format(path), shell=True,
                               stderr=STDOUT)
        except:
            try:
                # Try to delete in Unicode
                name = None
                from conda._vendor.auxlib.compat import Utf8NamedTemporaryFile
                from conda.utils import quote_for_shell

                with Utf8NamedTemporaryFile(mode="w", suffix=".bat", delete=False) as batch_file:
                    batch_file.write('RD /S {}\n'.format(quote_for_shell([path])))
                    batch_file.write('chcp 65001\n')
                    batch_file.write('RD /S {}\n'.format(quote_for_shell([path])))
                    batch_file.write('EXIT 0\n')
                    name = batch_file.name
                # If the above is bugged we can end up deleting hard-drives, so we check
                # that 'path' appears in it. This is not bulletproof but it could save you (me).
                with open(name, 'r') as contents:
                    content = contents.read()
                    assert path in content
                comspec = environ['COMSPEC']
                CREATE_NO_WINDOW = 0x08000000
                # It is essential that we `pass stdout=None, stderr=None, stdin=None` here because
                # if we do not, then the standard console handles get attached and chcp affects the
                # parent process (and any which share those console handles!)
                out = check_output([comspec, '/d', '/c', name], shell=False,
                                   stdout=None, stderr=None, stdin=None,
                                   creationflags=CREATE_NO_WINDOW)

            except CalledProcessError as e:
                pass
    else:
        try:
            makedirs('.empty')
        except:
            pass
        # yes, this looks strange.  See
        #    https://unix.stackexchange.com/a/79656/34459
        #    https://web.archive.org/web/20130929001850/http://linuxnote.net/jianingy/en/linux/a-fast-way-to-remove-huge-number-of-files.html  # NOQA
        rsync = which('rsync')
        if rsync and isdir('.empty'):
            try:
                out = check_output(
                    [rsync, '-a', '--delete', join(getcwd(), '.empty') + "/", path + "/"],
                    stderr=STDOUT)
            except CalledProcessError:
                pass
            shutil.rmtree('.empty')
    shutil.rmtree(path)


def unlink_or_rename_to_trash(path):
    """If files are in use, especially on windows, we can't remove them.
    The fallback path is to rename them (but keep their folder the same),
    which maintains the file handle validity.  See comments at:
    https://serverfault.com/a/503769
    """
    try:
        unlink(path)
    except EnvironmentError:
        try:
            rename(path, path + ".conda_trash")
        except EnvironmentError:
            pass


def remove_empty_parent_paths(path):
    # recurse to clean up empty folders that were created to have a nested hierarchy
    parent_path = dirname(path)
    while(isdir(parent_path) and not listdir(parent_path)):
        rmdir(parent_path)
        parent_path = dirname(parent_path)


def rm_rf(path, max_retries=5, trash=True, clean_empty_parents=False, *args, **kw):
    """
    Completely delete path
    max_retries is the number of times to retry on failure. The default is 5. This only applies
    to deleting a directory.
    If removing path fails and trash is True, files will be moved to the trash directory.
    """
    try:
        path = abspath(path)
        if isdir(path) and not islink(path):
            rmtree(path)
        elif lexists(path):
            unlink_or_rename_to_trash(path)
    finally:
        if lexists(path):
            return False
    if clean_empty_parents:
        remove_empty_parent_paths(path)
        return True



# -*- coding: utf-8 -*-
from __future__ import print_function

from conda_verify.checks import CondaPackageCheck, CondaRecipeCheck
from conda_verify.errors import PackageError, RecipeError
from conda_verify.utilities import ensure_list
from logging import getLogger


class Verify(object):
    """Verify class is called by the CLI but may be used as an API as well."""

    @staticmethod
    def verify_package(
        path_to_package=None, checks_to_ignore=None, exit_on_error=False, **kw
    ):
        """Run all package checks in order to verify a conda package.
        checks_to_ignore should be a list, tuple, or set of codes, such as ['C1102', 'C1104'].
        Codes are listed in readme.md.  Package codes follow 1xxx, recipe codes follow 2xxx."""
        package_check = CondaPackageCheck(path_to_package)

        if ("ignore_scripts" in kw and kw["ignore_scripts"]) or (
            "run_scripts" in kw and kw["run_scripts"]
        ):
            getLogger(__name__).warn(
                "Ignoring legacy ignore_scripts or run_scripts.  These have "
                "been replaced by the checks_to_ignore argument, which takes a"
                "list of codes, documented at https://github.com/conda/conda-verify#checks"
            )

        # collect all CondaPackageCheck methods that start with the word 'check'
        # this should later be a decorator that is placed on each check
        checks_to_display = []
        for method in dir(package_check):
            if method.startswith("check"):
                # runs the check
                #  TODO: should have a way to skip checks if a check's codes are all ignored
                check = getattr(package_check, method)()
                if check is not None and check.code not in ensure_list(
                    checks_to_ignore
                ):
                    checks_to_display.append(check)

        if checks_to_display and exit_on_error:
            raise PackageError(check)
        return (
            path_to_package,
            sorted(["[{}] {}".format(*c[1:]) for c in checks_to_display]),
        )

    @staticmethod
    def verify_recipe(
        rendered_meta=None,
        recipe_dir=None,
        checks_to_ignore=None,
        exit_on_error=False,
        **kw
    ):
        """Run all recipe checks in order to verify a conda recipe.
        checks_to_ignore should be a list, tuple, or set of codes, such as ['C2102', 'C2104'].
        Codes are listed in readme.md.  Package codes follow 1xxx, recipe codes follow 2xxx."""
        recipe_check = CondaRecipeCheck(rendered_meta, recipe_dir)

        if ("ignore_scripts" in kw and kw["ignore_scripts"]) or (
            "run_scripts" in kw and kw["run_scripts"]
        ):
            getLogger(__name__).warn(
                "Ignoring legacy ignore_scripts or run_scripts.  These have "
                "been replaced by the checks_to_ignore argument, which takes a"
                "list of codes, documented at https://github.com/conda/conda-verify#checks"
            )

        checks_to_display = []
        for method in (_ for _ in dir(recipe_check) if _.startswith("check_")):
            check = getattr(recipe_check, method)()
            if check and check.code not in ensure_list(checks_to_ignore):
                checks_to_display.append(check)

        if checks_to_display and exit_on_error:
            raise RecipeError(checks_to_display[0])
        return (
            recipe_dir,
            sorted(["[{}] {}".format(*c[1:]) for c in checks_to_display]),
        )



import json
import os
import sys
from glob import glob

import click
import tqdm
from concurrent.futures import ProcessPoolExecutor, as_completed

from conda_verify import __version__
from conda_verify.verify import Verify
from conda_verify.utilities import DummyExecutor, render_metadata, iter_cfgs


def _submit_verify_recipe(path, executor, ignore):
    futures = []
    for cfg in iter_cfgs():
        meta = render_metadata(path, cfg)
        if meta.get("build", {}).get("skip", "").lower() != "true":
            futures.append(
                executor.submit(
                    Verify.verify_recipe,
                    rendered_meta=meta,
                    recipe_dir=path,
                    checks_to_ignore=ignore,
                    exit_on_error=False,
                )
            )
    return futures


def _submit_verify_package(path, ignore):
    package_issues = (path, None)
    try:
        package_issues = Verify.verify_package(
            path_to_package=path, checks_to_ignore=ignore, exit_on_error=False
        )
    except (KeyError, OSError) as e:
        package_issues = (path, [str(e)])
    return package_issues


@click.command()
@click.argument("paths", nargs=-1, type=str)
@click.option("--ignore", nargs=1, type=str)
@click.option("--exit", is_flag=True)
@click.option("--debug", is_flag=True)
@click.option("--out-file", nargs=1, type=click.Path())
@click.version_option(prog_name="conda-verify", version=__version__)
def cli(paths, ignore, exit, debug, out_file):
    """conda-verify is a tool for validating conda packages and recipes.

    To validate a package:\n
    $  conda-verify path/to/package.tar.bz2

    To validate a recipe:\n
    $  conda-verify path/to/recipe_directory/
    """
    if ignore:
        ignore = ignore.split(",")

    package_issues = {}
    futures = []
    paths_glob = []
    for path in paths:
        glob_paths = glob(os.path.expanduser(path))
        if not glob_paths:
            print("Error: path spec %s didn't match any files" % path)
            sys.exit(1)
        paths_glob.extend(glob_paths)
    with (DummyExecutor if debug else ProcessPoolExecutor)() as executor:
        for path in paths_glob:
            meta_file = os.path.join(path, "meta.yaml")
            if os.path.isfile(meta_file):
                futures.extend(_submit_verify_recipe(path, executor, ignore))
            elif path.endswith((".tar.bz2", ".tar", ".conda")):
                futures.append(executor.submit(_submit_verify_package, path, ignore))
        for f in tqdm.tqdm(as_completed(futures), total=len(futures), leave=False):
            path, issues = f.result()
            if issues:
                package_issues[path] = issues

    if out_file:
        with open(out_file, "w") as f:
            json.dump(package_issues, f)
            print("saved to %s" % out_file)
    else:
        for path, issues in package_issues.items():
            print("-" * len(path))
            print(path)
            print("-" * len(path))
            for check in sorted(issues):
                try:
                    print(check, file=sys.stderr)
                except UnicodeEncodeError:
                    print(
                        "Could not print message for error code {} due to unicode error".format(
                            check.code
                        ),
                        file=sys.stderr,
                    )

    if exit and package_issues:
        sys.exit(1)



try:
    from conda_build.license_family import allowed_license_families as LICENSE_FAMILIES
except ImportError:
    print(
        "warning: could not import conda-build ALLOWED_LICENSE_FAMILIES data.  Falling back to "
        "possibly stale static list"
    )
    LICENSE_FAMILIES = [
        "AGPL",
        "GPL",
        "GPL2",
        "GPL3",
        "LGPL",
        "BSD",
        "MIT",
        "APACHE",
        "PSF",
        "CC",
        "PUBLIC-DOMAIN",
        "PROPRIETARY",
        "OTHER",
        "NONE",
    ]

try:
    from conda_build.metadata import FIELDS
except ImportError:
    print(
        "warning: could not import conda-build FIELDS data.  Falling back to possibly stale"
        "static list"
    )
    FIELDS = {
        "package": {"name", "version"},
        "source": {
            "fn",
            "url",
            "md5",
            "sha1",
            "sha256",
            "git_url",
            "git_tag",
            "git_branch",
            "git_rev",
            "patches",
            "hg_url",
            "hg_tag",
            "path",
        },
        "build": {
            "features",
            "track_features",
            "skip",
            "number",
            "entry_points",
            "osx_is_app",
            "noarch",
            "preserve_egg_dir",
            "win_has_prefix",
            "no_link",
            "ignore_prefix_files",
            "msvc_compiler",
            "skip_compile_pyc",
            "detect_binary_files_with_prefix",
            "script",
            "always_include_files",
            "binary_relocation",
            "binary_has_prefix_files",
            "noarch_python",
            "run_exports",
        },
        "requirements": {
            "build",
            "run",
            "preferred_env",
            "host",
            "preferred_env_executable_paths",
        },
        "outputs": {
            "name",
            "build",
            "about",
            "test",
            "version",
            "script",
            "requirements",
            "run_exports",
        },
        "app": {"entry", "icon", "summary", "type", "cli_opts"},
        "test": {"requires", "commands", "files", "source_files", "imports"},
        "about": {
            "license",
            "license_url",
            "license_family",
            "license_file",
            "summary",
            "description",
            "home",
            "doc_url",
            "doc_source_url",
            "dev_url",
        },
        "extra": {"recipe-maintainers", "final", "parent_recipe"},
    }

MAGIC_HEADERS = {
    b"\xca\xfe\xba\xbe": "MachO-universal",
    b"\xce\xfa\xed\xfe": "MachO-i386",
    b"\xcf\xfa\xed\xfe": "MachO-x86_64",
    b"\xfe\xed\xfa\xce": "MachO-ppc",
    b"\xfe\xed\xfa\xcf": "MachO-ppc64",
    b"MZ\x90\x00": "DLL",
    b"\x7fELF": "ELF",
}

DLL_TYPES = {
    0x0: "UNKNOWN",
    0x1D3: "AM33",
    0x8664: "AMD64",
    0x1C0: "ARM",
    0xEBC: "EBC",
    0x14C: "I386",
    0x200: "IA64",
    0x9041: "M32R",
    0x266: "MIPS16",
    0x366: "MIPSFPU",
    0x466: "MIPSFPU16",
    0x1F0: "POWERPC",
    0x1F1: "POWERPCFP",
    0x166: "R4000",
    0x1A2: "SH3",
    0x1A3: "SH3DSP",
    0x1A6: "SH4",
    0x1A8: "SH5",
    0x1C2: "THUMB",
    0x169: "WCEMIPSV2",
}

CONDA_FORGE_COMMENTS = """
# Note: there are many handy hints in comments in this example -- remove them when you've finalized your recipe
# Jinja variables help maintain the recipe as you'll update the version only here.
# sha256 is the prefered checksum -- you can get it for a file with:
#  `openssl sha256 <file name>`.
# You may need the openssl package, available on conda-forge
#  `conda install openssl -c conda-forge``
# If the installation is complex, or different between Unix and Windows, use separate bld.bat and build.sh files instead of this key.
# By default, the package will be built for the Python versions supported by conda-forge and for all major OSs.
# Add the line "skip: True  # [py<35]" (for example) to limit to Python 3.5 and newer, or "skip: True  # [not win]" to limit to Windows.
# When setuptools is available add the `--single-version-externally-managed --record record.txt` above.
# if your project compiles code (such as a C extension) then add `toolchain` as a build requirement.
# Some package might need a `test/commands` key to check CLI.
# List all the packages/modules that `run_test.py` imports.
# Remember to specify the license variants for BSD, Apache, GPL, and LGLP.
# Prefer the short version, e.g: GPL-2.0 instead of GNU General Public License version 2.0
# See https://opensource.org/licenses/alphabetical
# The license_family, i.e. "BSD" if license is "BSD-3-Clause". (optional)
# It is strongly encouraged to include a license file in the package,
# (even if the license doesn't require it) using the license_file entry.
# See http://conda.pydata.org/docs/building/meta-yaml.html#license-file
# The remaining entries in this section are optional, but recommended
# GitHub IDs for maintainers of the recipe.
# Always check with the people listed below if they are OK becoming maintainers of the recipe. (There will be spam!)
"""
